/*  tse.c - VS Bomber: Tile and Sprite Engine
 *
 *  Copyright 1998, Benjamin C. W. Sittler and Stefan M. W. Sittler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "game.h"
#include "tse.h"

/* video mode */
int screen_w = 640, screen_h = 480, screen_depth = 8;

/* character size */
#ifndef char_w
int char_w = 32;
#endif
#ifndef char_h
int char_h = 32;
#endif

/* character set memory bitmaps */
BITMAP *charset[CHARSET_COLS * CHARSET_ROWS] = { 0, };

/* renderers */
int tse_style = TSE_GOOD;

/* screen matrix */
struct Cell vs_screen[VS_SCREENS][ROWS][COLS] = { { { {0, }, }, }, };

/* sprite table */
struct Sprite sprite[SPRITES] = { {0, }, };

/* sprite priority table */
int sprite_prio[SPRITES];

/* pixel offsets of screen matrix from viewport's upper left corner */
int x_offset, y_offset;

/* dirty the region occupied by a sprite, re-sort it in the priority table */
void dirty_sprite(int s) {
   int x, y, r, c, h, w, r1, c1, s1, prio, z, z1;

   /* re-prioritize sprite */
   if (sprite[s] . used) {
     while ((prio = sprite[s] . prio) &&
	    (((z1 = sprite[(s1 = sprite_prio[prio - 1])] . z) <
	      (z = sprite[s] . z)) ||
	     ! sprite[s1] . used)) {
       sprite_prio[prio] = s1;
       sprite_prio[prio - 1] = s;
       sprite[s1] . prio = prio;
       sprite[s] . prio = -- prio;
     }
     while (((prio = sprite[s] . prio) + 1 < SPRITES) &&
	    ((z1 = sprite[(s1 = sprite_prio[prio + 1])] . z) >
	     (z = sprite[s] . z)) &&
	    sprite[s1] . used) {
       sprite_prio[prio] = s1;
       sprite_prio[prio + 1] = s;
       sprite[s1] . prio = prio;
       sprite[s] . prio = ++ prio;
     }
   } else
      while (((prio = sprite[s] . prio) + 1 < SPRITES) &&
	     sprite[(s1 = sprite_prio[prio + 1])] . used) {
         sprite_prio[prio] = s1;
	 sprite_prio[prio + 1] = s;
	 sprite[s1] . prio = prio;
	 sprite[s] . prio = ++ prio;
      }
   /* mark sprite's region dirty */
   x = sprite[s] . x;
   y = sprite[s] . y;
   c1 = (x + sprite[s] . c * char_w) / char_w;
   if (c1 < 0)
     c1 = 0;
   r1 = (y + sprite[s] . r * char_h) / char_h;
   if (r1 < 0)
     r1 = 0;
   h = sprite[s] . h + (y != 0) + r1;
   if (h >= ROWS)
     h = ROWS - 1;
   w = sprite[s] . w + (x != 0) + c1;
   if (w >= COLS)
     w = COLS - 1;
   for (r = r1; r < h; r ++)
     for (c = c1; c < w; c ++)
       vs_screen[0][r][c] . dirty = 1;
}

/* updates a sprite's frame timer, returns 1 when the sprite loops */
/* if dirty is NULL, dirty_sprite(s) is called; otherwise *dirty is set */
/* A B C B, A B C B, ... */
int sprite_loop(int s, int *dirty) {
   if (sprite[s] . timer &&
       ! -- sprite[s] . timer) {
      sprite[s] . timer = sprite[s] . delay;
      if (sprite[s] . lw > 1) {
         if (! dirty)
            dirty_sprite(s);
         else
            *dirty = 1;
         sprite[s] . fr = (sprite[s] . fr + 1) %
                          ((sprite[s] . lw - 1) * 2);
      }
      if (! sprite[s] . fr)
         return 1;
   }
   return 0;
}


/* updates a sprite's frame timer, returns 1 when the sprite repeats */
/* if dirty is NULL, dirty_sprite(s) is called; otherwise *dirty is set */
/* A B C, A B C, ... */
int sprite_repeat(int s, int *dirty) {
   if (sprite[s] . timer &&
       ! -- sprite[s] . timer) {
      sprite[s] . timer = sprite[s] . delay;
      if (sprite[s] . lw > 1) {
         if (! dirty)
            dirty_sprite(s);
         else
            *dirty = 1;
         sprite[s] . fr = (sprite[s] . fr + 1) % sprite[s] . lw;
      }
      if (! sprite[s] . fr)
         return 1;
   }
   return 0;
}

void update_good(BITMAP *bmp, int upper);
void update_fast(int upper);
void update_ripple(BITMAP *bmp, int upper);

/* ripple scanline in characters */
int ripple = 0, ripple_h = 1;

/* update screen from screen matrices and sprite table */
void update_screen(BITMAP *bmp, int upper) {
  switch (tse_style) {
  case TSE_GOOD:
    update_good(bmp, upper);
    break;
  case TSE_FAST:
    update_fast(upper);
    break;
  case TSE_VSYNC:
    vsync();
    update_fast(upper);
    break;
  case TSE_RIPPLE:
    update_ripple(bmp, upper);
    break;
  }
}

/* perfect (but somewhat slow) update */
void update_good(BITMAP *bmp, int upper) {
  int r, c, x, y, s, s1, w, h, ch, lw, fr, st;
  
  /* update dirty tiles */
  for (r = 0, y = 0; r < ROWS; r ++, y += char_h)
    for (c = 0, x = 0; c < COLS; c ++, x += char_w)
      if (vs_screen[0][r][c] . dirty) {
	/* draw lower screen */
	vs_screen[0][r][c] . dirty = 0;
	lw = vs_screen[0][r][c] . lw;
	fr = vs_screen[0][r][c] . fr;
	if (fr < lw)
	  ch = vs_screen[0][r][c] . ch + fr;
	else
	  ch = vs_screen[0][r][c] . ch + 2 * (lw - 1) - fr;
	if ((st = vs_screen[0][r][c] . st))
	  ch = CHARSET(st, ch);
	blit(charset[ch],
	     bmp,
	     0, 0,
	     0, 0,
	     char_w, char_h);
	/* mask in sprites */
	for (s1 = 0; s1 < SPRITES; s1 ++) {
	  s = sprite_prio[s1];
	  if (sprite[s] . used) {
	    int x1, y1, r1, c1;
	    
	    w = sprite[s] . w;
	    x1 = sprite[s] . c * char_w + sprite[s] . x - x;
	    if ((x1 >= char_w) || (x1 + w * char_w <= 0))
	      continue;
	    h = sprite[s] . h;
	    y1 = sprite[s] . r * char_h + sprite[s] . y - y;
	    if ((y1 >= char_h) || (y1 + h * char_h <= 0))
	      continue;
	    lw = sprite[s] . lw;
	    fr = sprite[s] . fr;
	    if (fr < lw)
	      ch = sprite[s] . ch + w * fr;
	    else
	      ch = sprite[s] . ch + w * (2 * (lw - 1) - fr);
	    st = sprite[s] . st;
	    if ((h == 1) && (w == 1))
	      draw_sprite(bmp,
			  charset[CHARSET(st, ch)],
			  x1, y1);
	    else {
	      int x2, y2;
	      
	      ch = CHARSET(h * st, ch);
	      for (r1 = 0, y2 = y1; r1 < h; r1 ++, y2 += char_h)
		for (c1 = 0, x2 = x1; c1 < w; c1 ++, x2 += char_w)
		  draw_sprite(bmp,
			      charset[CHARSET(r1, ch + c1)],
			      x2, y2);
	    }
	  }
	}
	if (upper && (ch = vs_screen[1][r][c] . ch)) {
	  /* draw upper screen */
	  lw = vs_screen[1][r][c] . lw;
	  fr = vs_screen[1][r][c] . fr;
	  st = vs_screen[1][r][c] . st;
	  if (fr < lw)
	    ch += fr;
	  else
	    ch += 2 * (lw - 1) - fr;
	  ch = CHARSET(st, ch);
	  draw_sprite(bmp,
		      charset[ch],
		      0, 0);
	}
	blit(bmp,
	     screen,
	     0, 0,
	     x + x_offset, y + y_offset,
	     char_w, char_h);
      }
}

/* complete update w/rippling */
void update_ripple(BITMAP *bmp, int upper) {
  int r, c, x, y, s, s1, w, h, ch, lw, fr, st;

  /* update all tiles */
  for (r = 0, y = 0; r < ROWS; r ++, y += char_h)
    for (c = 0, x = 0; c < COLS; c ++, x += char_w) {
      /* draw lower screen */
      lw = vs_screen[0][r][c] . lw;
      fr = vs_screen[0][r][c] . fr;
      if (fr < lw)
	ch = vs_screen[0][r][c] . ch + fr;
      else
	ch = vs_screen[0][r][c] . ch + 2 * (lw - 1) - fr;
      if ((st = vs_screen[0][r][c] . st))
	ch = CHARSET(st, ch);
      blit(charset[ch],
	   bmp,
	   0, 0,
	   0, 0,
	   char_w, char_h);
      if (ripple)
	blit(charset[ch],
	     bmp,
	     0, ripple - 1,
	     0, ripple,
	     char_w, ripple_h);
      /* mask in sprites */
      for (s1 = 0; s1 < SPRITES; s1 ++) {
	s = sprite_prio[s1];
	if (sprite[s] . used) {
	  int x1, y1, r1, c1;
	  
	  w = sprite[s] . w;
	  x1 = sprite[s] . c * char_w + sprite[s] . x - x;
	  if ((x1 >= char_w) || (x1 + w * char_w <= 0))
	    continue;
	  h = sprite[s] . h;
	  y1 = sprite[s] . r * char_h + sprite[s] . y - y;
	  if ((y1 >= char_h) || (y1 + h * char_h <= 0))
	    continue;
	  lw = sprite[s] . lw;
	  fr = sprite[s] . fr;
	  if (fr < lw)
	    ch = sprite[s] . ch + w * fr;
	  else
	    ch = sprite[s] . ch + w * (2 * (lw - 1) - fr);
	  st = sprite[s] . st;
	  if ((h == 1) && (w == 1))
	    draw_sprite(bmp,
			charset[CHARSET(st, ch)],
			x1, y1);
	  else {
	    int x2, y2;
	    
	    ch = CHARSET(h * st, ch);
	    for (r1 = 0, y2 = y1; r1 < h; r1 ++, y2 += char_h)
	      for (c1 = 0, x2 = x1; c1 < w; c1 ++, x2 += char_w)
		draw_sprite(bmp,
			    charset[CHARSET(r1, ch + c1)],
			    x2, y2);
	  }
	}
      }
      if (upper && (ch = vs_screen[1][r][c] . ch)) {
	/* draw upper screen */
	lw = vs_screen[1][r][c] . lw;
	fr = vs_screen[1][r][c] . fr;
	st = vs_screen[1][r][c] . st;
	if (fr < lw)
	  ch += fr;
	else
	  ch += 2 * (lw - 1) - fr;
	ch = CHARSET(st, ch);
	draw_sprite(bmp,
		    charset[ch],
		    0, 0);
      }
      blit(bmp,
	   screen,
	   0, 0,
	   x + x_offset, y + y_offset,
	   char_w, char_h);
    }
}

/* imperfect (but somewhat speedy) update */
void update_fast(int upper) {
  int r, c, r1, c1, x, y, x1, y1, s, s1, w, h, ch, lw, fr, st;
  
  /* update dirty tiles in lower screen */
  for (r = 0, y = y_offset; r < ROWS; r ++, y += char_h)
    for (c = 0, x = x_offset; c < COLS; c ++, x += char_w)
      if (vs_screen[0][r][c] . dirty) {
	/* draw lower screen */
	lw = vs_screen[0][r][c] . lw;
	fr = vs_screen[0][r][c] . fr;
	if (fr < lw)
	  ch = vs_screen[0][r][c] . ch + fr;
	else
	  ch = vs_screen[0][r][c] . ch + 2 * (lw - 1) - fr;
	if ((st = vs_screen[0][r][c] . st))
	  ch = CHARSET(st, ch);
	blit(charset[ch],
	     screen,
	     0, 0,
	     x, y,
	     char_w, char_h);
      }
  /* mask in sprites */
  for (s1 = 0; s1 < SPRITES; s1 ++) {
    s = sprite_prio[s1];
    if (sprite[s] . used) {
      w = sprite[s] . w;
      x = sprite[s] . c * char_w + sprite[s] . x + x_offset;
      h = sprite[s] . h;
      y = sprite[s] . r * char_h + sprite[s] . y + y_offset;
      lw = sprite[s] . lw;
      fr = sprite[s] . fr;
      if (fr < lw)
	ch = sprite[s] . ch + w * fr;
      else
	ch = sprite[s] . ch + w * (2 * (lw - 1) - fr);
      st = sprite[s] . st;
      if ((h == 1) && (w == 1))
	draw_sprite(screen,
		    charset[CHARSET(st, ch)],
		    x, y);
      else {
	ch = CHARSET(h * st, ch);
	for (r1 = 0, y1 = y; r1 < h; r1 ++, y1 += char_h)
	  for (c1 = 0, x1 = x; c1 < w; c1 ++, x1 += char_w) {
	    draw_sprite(screen,
			charset[CHARSET(r1, ch + c1)],
			x1, y1);
	  }
      }
      if (upper)
	dirty_sprite(s);
    }
  }
  /* update dirty tiles in upper screen */
  for (r = 0, y = y_offset; r < ROWS; r ++, y += char_h)
    for (c = 0, x = x_offset; c < COLS; c ++, x += char_w)
      if (vs_screen[0][r][c] . dirty) {
	vs_screen[0][r][c] . dirty = 0;
	if (upper && (ch = vs_screen[1][r][c] . ch)) {
	  lw = vs_screen[1][r][c] . lw;
	  fr = vs_screen[1][r][c] . fr;
	  st = vs_screen[1][r][c] . st;
	  if (fr < lw)
	    ch += fr;
	  else
	    ch += 2 * (lw - 1) - fr;
	  ch = CHARSET(st, ch);
	  draw_sprite(screen,
		      charset[ch],
		      x, y);
	}
      }
}
